@@ -19,7 +19,7 @@
 #if !(defined(SPARC) || defined(POWER) || defined(POWERPC) || defined(M88000) \
       || defined(AM29K) || defined(MC68000) || defined(MC68020) \
       || defined(I386) || defined(PYR) || defined(ALPHA) || defined(HPPA) \
-      || defined(SH))
+      || defined(SH) || defined (I960) || defined (I960B))
 /* If no target instruction set is defined, use host instruction set.  */
 #define SPARC (defined(sparc) || defined(__sparc__))
 #define POWER ((defined(rs6000) || defined(_IBMR2)) && !defined (_ARCH_PPC))
@@ -34,13 +34,19 @@
 #define ALPHA defined(__alpha)
 #define HPPA defined(__hppa)
 #define SH defined(__sh__)
+#define I960 defined (__i960)
 #endif
 
 #define M68000 (MC68000 || MC68020)
-#if defined(POWERPC)
+
+#if POWERPC
 #define POWER 1
 #endif
 
+#if I960B
+#define I960 1
+#endif
+
 #if SPARC
 #define TARGET_STRING "SPARC v7/v8"
 #elif POWERPC
@@ -64,27 +70,33 @@
 #elif HPPA
 #define TARGET_STRING "Hewlett-Packard Precision Architecture (PA-RISC)"
 #elif  SH
-#define TARGET_STRING "Hitatchi Super-H (SH)"
+#define TARGET_STRING "Hitachi Super-H (SH)"
+/* Reject sequences that require two different registers to be allocated to
+   register r0.  */
 #define EXTRA_SEQUENCE_TESTS(seq, n) \
 {									\
-  int reg0 = -1;							\
+  int reg0 = -1;	/* -1 means r0 is not yet allocated */		\
   int i;								\
   for (i = 0; i < n; i++)						\
     {									\
       if ((seq[i].opcode == AND || seq[i].opcode == XOR			\
 	   || seq[i].opcode == IOR || seq[i].opcode == CYEQ)		\
-	  && IMMEDIATE_P (seq[i].s2))					\
+	  && IMMEDIATE_P (seq[i].s2) && IMMEDIATE_VAL (seq[i].s2) != 0)	\
 	{								\
-	  if (reg0 >= 0)						\
+	  if (reg0 >= 0 && reg0 != seq[i].d)				\
 	    return;							\
-	  reg0 = i;							\
+	  reg0 = seq[i].d;						\
 	}								\
     }									\
 }
+#elif  I960
+#define TARGET_STRING "Intel 960 v1.0"
+#elif  I960B
+#define TARGET_STRING "Intel 960 v1.1"
 #endif
 
 #if !(SPARC || POWER || M88000 || AM29K || M68000 || I386 || PYR \
-      || ALPHA || HPPA || SH)
+      || ALPHA || HPPA || SH || I960)
 #error You have to choose target CPU type (e.g. -DSPARC).
 #endif
 
@@ -94,39 +106,41 @@
 #define SHIFT_COST(CNT) 1
 #endif
 
-#ifdef SHIFTS
-#undef SHIFTS
-#define SHIFTS 1
-#else
-#define SHIFTS 0
-#endif
-
-#ifndef USE_ASM
-#define NO_ASM 1		/* Used by longlong.h */
-#endif
-
-#include "longlong.h"
-
 #if ALPHA
 #define BITS_PER_WORD 64
-#else /* ! ALPHA */
+#else
 #define BITS_PER_WORD 32
 #endif
 
+/* Get longlong.h for double-word arithmetic.
+   First define the types for it to operate on.  */
+#define UWtype word
+#define UHWtype word
+#define UDWtype unsigned long	/* Bogus, but we'll not depend on it */
+#define W_TYPE_SIZE BITS_PER_WORD
+#define SItype int
+#define USItype unsigned int
+#define DItype long long int
+#define UDItype unsigned long long int
+#define LONGLONG_STANDALONE
+#include "longlong.h"
+
 #if HPPA
 #define HAS_NULLIFICATION 1
 enum { NOT_NULLIFY = 0, NULLIFY = 1 };
 #endif
 
 #if BITS_PER_WORD == 64
-#ifdef __GNUC__
+#if defined (__GNUC__) || defined (_LONGLONG)
 typedef unsigned long long int unsigned_word;
 typedef signed long long int signed_word;
 typedef unsigned_word word;
+#undef PSTR			/* no portable way to print */
 #elif __alpha /* Native compiler on alpha has 64 bit longs.  */
 typedef unsigned long int unsigned_word;
 typedef signed long int signed_word;
 typedef unsigned_word word;
+#define PSTR "0x%lx"
 #else /* Not on alpha, not GCC.  Don't have 64 bit type.  */
 #error Do not know how to perform 64 bit arithmetic with this compiler.
 #endif
@@ -134,16 +148,17 @@ typedef unsigned_word word;
 typedef unsigned int unsigned_word;
 typedef signed int signed_word;
 typedef unsigned_word word;
+#define PSTR "0x%x"
 #endif
 
 
-#define TRUNC_CNT(cnt) ((cnt) & (BITS_PER_WORD - 1))
+#define TRUNC_CNT(cnt) ((unsigned) (cnt) % BITS_PER_WORD)
 
 #if defined(sparc) || defined(__GNUC__)
 #define alloca __builtin_alloca
 #endif
 
-#if !defined(__GNUC__) || !defined(__OPTIMIZE__)
+#if !defined(__GNUC__) || !defined(__OPTIMIZE__) || defined(DEBUG)
 #define inline /* Empty */
 #endif
 
@@ -162,7 +177,7 @@ typedef unsigned_word word;
 #define CNST(n) (0x20 + n)
 #define VALUE(n) n
 
-/* The IMMEDIATE_* macros are for printing assembler.  NOT for sequence
+/* The IMMEDIATE_* macros are for printing assembly.  NOT for sequence
    generating or analyze.  */
 #define IMMEDIATE_P(op) (op >= 0x20 - 5)
 static const word __immediate_val[] =
@@ -504,11 +519,11 @@ typedef struct
 #endif
 #ifndef PERFORM_SHIFTL
 #define PERFORM_SHIFTL(d, co, r1, r2, ci) \
-  ((d) = ((signed_word) (r1) << TRUNC_CNT(r2)), (co) = (ci))
+  ((d) = ((unsigned_word) (r1) << TRUNC_CNT(r2)), (co) = (ci))
 #endif
 #ifndef PERFORM_ROTATEL
 #define PERFORM_ROTATEL(d, co, r1, r2, ci) \
- ((d) = (r2) == 0 ? (r1)						\
+ ((d) = TRUNC_CNT(r2) == 0 ? (r1)					\
   : ((r1) << TRUNC_CNT(r2)) | ((r1) >> TRUNC_CNT(BITS_PER_WORD - (r2))),\
   (co) = (ci))
 #endif
@@ -533,7 +548,7 @@ typedef struct
 #endif
 #ifndef PERFORM_SHIFTL_CO
 #define PERFORM_SHIFTL_CO(d, co, r1, r2, ci) \
-  do { word __d = ((signed_word) (r1) << TRUNC_CNT(r2));		\
+  do { word __d = ((unsigned_word) (r1) << TRUNC_CNT(r2));		\
        (co) = ((r1) >> TRUNC_CNT(BITS_PER_WORD - (r2))) & 1;		\
        (d) = __d; } while (0)
 #endif
@@ -716,11 +731,15 @@ typedef struct
 #ifndef PERFORM_CLZ
 #define PERFORM_CLZ(d, co, r1, ci) \
   do {									\
-    int __a;								\
-    __a = (r1) <= 0xffff						\
-      ? ((r1) <= 0xff ? 0 : 8)						\
-      : ((r1) <= 0xffffff ?  16 : 24);					\
-    (d) = clz_tab[(r1) >> __a] - __a;					\
+    int __a = 0;							\
+    word __r = (r1);							\
+    if (__r > 0xffffffff)						\
+      __r = __r >> 31 >> 1, __a += 32;					\
+    if (__r > 0xffff)							\
+      __r >>= 16, __a += 16;						\
+    if (__r > 0xff)							\
+      __r >>= 8, __a += 8;						\
+    (d) = clz_tab[__r] - __a + BITS_PER_WORD - 32;			\
     (co) = (ci);							\
   } while (0)
 #endif
@@ -729,6 +748,7 @@ typedef struct
 #define PERFORM_CTZ(d, co, r1, ci) \
   do {									\
     int __a;								\
+    abort ();								\
     __a = ((r1) & 0xffff == 0)						\
       ? (((r1) & 0xff0000) == 0 ? 24 : 16)				\
       : ((r1) & 0xff == 0) ? 8 : 0;					\
@@ -745,17 +765,8 @@ typedef struct
       : ((r1) <= 0xffffff ?  16 : 24);					\
     (d) = ff1_tab[(r1) >> __a] + __a;					\
     (co) = (ci);							\
-    FF1_CHECK(d,r1)							\
   } while (0)
 #endif
-#if m88k
-#define FF1_CHECK(d,r1)							\
-  { int t;								\
-    asm ("ff1 %0,%1" : "=r" (t) : "r" (r1));				\
-    if (t != (d)) abort (); }
-#else
-#define FF1_CHECK(d,r1)
-#endif
 #ifndef PERFORM_FF0
 #define PERFORM_FF0(d, co, r1, ci) \
   PERFORM_FF1(d, co, ~(r1), ci)
@@ -765,7 +776,7 @@ typedef struct
   do {									\
     word __x = (r1) & (-r1);						\
     PERFORM_CLZ(d, co, __x, ci);					\
-    (d) = 32 - (d);							\
+    (d) = BITS_PER_WORD - (d);						\
   } while (0)
 #endif
 #ifndef PERFORM_BSF86
@@ -861,6 +872,8 @@ typedef struct
 /* HP-PA nullifying instructions.  */
 #define PERFORM_NULLIFIED(v, co, sc, ci) \
   ((v) = values[dr], (co) = (ci), (sc) = 0)
+#define PERFORM_COPY_S(d, co, sc, r1, ci) \
+  ((d) = (r1), (co) = (ci), (sc) = 1)
 
 #define PERFORM_ADD_SEQ(d, co, sc, r1, r2, ci) \
   ((d) = (r1) + (r2), (co) = (ci),					\
@@ -1294,28 +1307,134 @@ typedef struct
 #define PERFORM_ANDC_S(d, co, sc, r1, r2, ci) \
   ((d) = (r1) & ~(r2), (co) = (ci), (sc) = 1)
 
-#if LATER
-#define PERFORM_LSHIFTR(d, co, sc, r1, r2, ci) \
-  ((d) = ((unsigned_word) (r1) >> TRUNC_CNT(r2)),			\
-   (co) = (ci))
-#define PERFORM_ASHIFTR(d, co, sc, r1, r2, ci) \
-  ((d) = ((signed_word) (r1) >> TRUNC_CNT(r2)),				\
-   (co) = (ci))
-#define PERFORM_SHIFTL(d, co, sc, r1, r2, ci) \
-  ((d) = ((signed_word) (r1) << TRUNC_CNT(r2)), (co) = (ci))
-#define PERFORM_ROTATEL(d, co, sc, r1, r2, ci) \
- ((d) = (r2) == 0 ? (r1)						\
+#define PERFORM_LSHIFTR_S(d, co, sc, r1, r2, ci) \
+  ((d) = ((unsigned_word) (r1) >> TRUNC_CNT(r2)), (co) = (ci), (sc) = 1)
+#define PERFORM_ASHIFTR_S(d, co, sc, r1, r2, ci) \
+  ((d) = ((signed_word) (r1) >> TRUNC_CNT(r2)), (co) = (ci), (sc) = 1)
+#define PERFORM_SHIFTL_S(d, co, sc, r1, r2, ci) \
+  ((d) = ((unsigned_word) (r1) << TRUNC_CNT(r2)), (co) = (ci), (sc) = 1)
+#define PERFORM_ROTATEL_S(d, co, sc, r1, r2, ci) \
+ ((d) = TRUNC_CNT(r2) == 0 ? (r1)					\
   : ((r1) << TRUNC_CNT(r2)) | ((r1) >> TRUNC_CNT(BITS_PER_WORD - (r2))),\
-  (co) = (ci))
-#define PERFORM_EXTS1(d, co, sc, r1, r2, ci) \
-  ((d) = ((signed_word) (r1) >> TRUNC_CNT(r2)) << 31 >> 31, (co) = (ci))
-#define PERFORM_EXTS2(d, co, sc, r1, r2, ci) \
-  ((d) = ((signed_word) (r1) >> TRUNC_CNT(r2)) << 30 >> 30, (co) = (ci))
-#define PERFORM_EXTU1(d, co, sc, r1, r2, ci) \
-  ((d) = ((unsigned_word) (r1) >> TRUNC_CNT(r2)) & 1, (co) = (ci))
-#define PERFORM_EXTU2(d, co, sc, r1, r2, ci) \
-  ((d) = ((unsigned_word) (r1) >> TRUNC_CNT(r2)) & 3, (co) = (ci))
-#endif
+  (co) = (ci), (sc) = 1)
+#define PERFORM_EXTS1_S(d, co, sc, r1, r2, ci) \
+  ((d) = ((signed_word) (r1) >> TRUNC_CNT(r2)) << 31 >> 31, (co) = (ci), (sc) = 1)
+#define PERFORM_EXTS2_S(d, co, sc, r1, r2, ci) \
+  ((d) = ((signed_word) (r1) >> TRUNC_CNT(r2)) << 30 >> 30, (co) = (ci), (sc) = 1)
+#define PERFORM_EXTS8_S(d, co, sc, r1, r2, ci) \
+  ((d) = ((signed_word) (r1) >> TRUNC_CNT(r2)) << 24 >> 24, (co) = (ci), (sc) = 1)
+#define PERFORM_EXTS16_S(d, co, sc, r1, r2, ci) \
+  ((d) = ((signed_word) (r1) >> TRUNC_CNT(r2)) << 16 >> 16, (co) = (ci), (sc) = 1)
+#define PERFORM_EXTU1_S(d, co, sc, r1, r2, ci) \
+  ((d) = ((unsigned_word) (r1) >> TRUNC_CNT(r2)) & 1, (co) = (ci), (sc) = 1)
+#define PERFORM_EXTU2_S(d, co, sc, r1, r2, ci) \
+  ((d) = ((unsigned_word) (r1) >> TRUNC_CNT(r2)) & 3, (co) = (ci), (sc) = 1)
+
+#define IB0(x) x
+#define IB1(x) ((x) << 1)
+#define IB2(x) ((x) << 2)
+#define EB0(x) ((x) & 1)
+#define EB1(x) (((x) >> 1) & 1)
+#define EB2(x) (((x) >> 2) & 1)
+
+#define PERFORM_ADDC_960(d, co, r1, r2, ci) \
+  do { word __d = (r1) + EB1(ci);					\
+       word __cy = __d < (r1);						\
+       (d) = __d + (r2);						\
+       (co) = (IB1(((d) < __d) + __cy)					\
+	       | IB0((signed_word) (~((r1) ^ (r2)) & ((d) ^ (r1))) < 0));\
+  } while (0)
+#define PERFORM_SUBC_960(d, co, r1, r2, ci) \
+  do { word __d = (r1) + EB1(ci);					\
+       word __cy = __d < (r1);						\
+       (d) = __d + ~(r2);						\
+       (co) = (IB1(((d) < __d) + __cy)					\
+	       | IB0((signed_word) (~((r1) ^ ~(r2)) & ((d) ^ (r1))) < 0));\
+  } while (0)
+#define PERFORM_SEL_NO_960(d, co, r1, r2, ci) \
+  ((d) = (ci) == 0 ? (r2) : (r1), (co) = (ci))
+#define PERFORM_SEL_G_960(d, co, r1, r2, ci) \
+  ((d) = ((ci) & 1) != 0 ? (r2) : (r1), (co) = (ci))
+#define PERFORM_SEL_E_960(d, co, r1, r2, ci) \
+  ((d) = ((ci) & 2) != 0 ? (r2) : (r1), (co) = (ci))
+#define PERFORM_SEL_GE_960(d, co, r1, r2, ci) \
+  ((d) = ((ci) & 3) != 0 ? (r2) : (r1), (co) = (ci))
+#define PERFORM_SEL_L_960(d, co, r1, r2, ci) \
+  ((d) = ((ci) & 4) != 0 ? (r2) : (r1), (co) = (ci))
+#define PERFORM_SEL_NE_960(d, co, r1, r2, ci) \
+  ((d) = ((ci) & 5) != 0 ? (r2) : (r1), (co) = (ci))
+#define PERFORM_SEL_LE_960(d, co, r1, r2, ci) \
+  ((d) = ((ci) & 6) != 0 ? (r2) : (r1), (co) = (ci))
+#define PERFORM_SEL_O_960(d, co, r1, r2, ci) \
+  ((d) = ((ci) & 7) != 0 ? (r2) : (r1), (co) = (ci))
+#define PERFORM_CONCMPO_960(d, co, r1, r2, ci) \
+  ((co) = EB2(ci) == 0							\
+   ? ((unsigned_word) (r1) <= (unsigned_word) (r2)) + 1 : (ci))
+#define PERFORM_CONCMPI_960(d, co, r1, r2, ci) \
+  ((co) = EB2(ci) == 0							\
+   ? ((signed_word) (r1) <= (signed_word) (r2)) + 1 : (ci))
+#define PERFORM_CMPO_960(d, co, r1, r2, ci) \
+  ((co) = (IB0((unsigned_word) (r1) > (unsigned_word) (r2))		\
+	   | IB1((r1) == (r2))						\
+	   | IB2((unsigned_word) (r1) < (unsigned_word) (r2))))
+#define PERFORM_CMPI_960(d, co, r1, r2, ci) \
+  ((co) = (IB0((signed_word) (r1) > (signed_word) (r2))			\
+	   | IB1((r1) == (r2))						\
+	   | IB2((signed_word) (r1) < (signed_word) (r2))))
+#define PERFORM_SHIFTL_NT(d, co, r1, r2, ci) \
+  ((d) = ((r2) < BITS_PER_WORD ? ((unsigned_word) (r1) << (r2)) : 0), (co) = (ci))
+#define PERFORM_LSHIFTR_NT(d, co, r1, r2, ci) \
+  ((d) = ((r2) < BITS_PER_WORD ? ((unsigned_word) (r1) >> (r2)) : 0), (co) = (ci))
+#define PERFORM_ASHIFTR_NT(d, co, r1, r2, ci) \
+  ((d) = ((r2) < BITS_PER_WORD						\
+	  ? ((signed_word) (r1) >> (r2))				\
+	  : -(signed_word) (r1) < 0), (co) = (ci))
+#define PERFORM_ADDO_NO_960(d, co, r1, r2, ci) \
+  ((d) = (ci) == 0 ? (r1) + (r2) : (d), (co) = (ci))
+#define PERFORM_ADDO_G_960(d, co, r1, r2, ci) \
+  ((d) = ((ci) & 1) != 0 ? (r1) + (r2) : (d), (co) = (ci))
+#define PERFORM_ADDO_E_960(d, co, r1, r2, ci) \
+  ((d) = ((ci) & 2) != 0 ? (r1) + (r2) : (d), (co) = (ci))
+#define PERFORM_ADDO_GE_960(d, co, r1, r2, ci) \
+  ((d) = ((ci) & 3) != 0 ? (r1) + (r2) : (d), (co) = (ci))
+#define PERFORM_ADDO_L_960(d, co, r1, r2, ci) \
+  ((d) = ((ci) & 4) != 0 ? (r1) + (r2) : (d), (co) = (ci))
+#define PERFORM_ADDO_NE_960(d, co, r1, r2, ci) \
+  ((d) = ((ci) & 5) != 0 ? (r1) + (r2) : (d), (co) = (ci))
+#define PERFORM_ADDO_LE_960(d, co, r1, r2, ci) \
+  ((d) = ((ci) & 6) != 0 ? (r1) + (r2) : (d), (co) = (ci))
+#define PERFORM_ADDO_O_960(d, co, r1, r2, ci) \
+  ((d) = ((ci) & 7) != 0 ? (r1) + (r2) : (d), (co) = (ci))
+
+#define PERFORM_SUBO_NO_960(d, co, r1, r2, ci) \
+  ((d) = (ci) == 0 ? (r1) - (r2) : (d), (co) = (ci))
+#define PERFORM_SUBO_G_960(d, co, r1, r2, ci) \
+  ((d) = ((ci) & 1) != 0 ? (r1) - (r2) : (d), (co) = (ci))
+#define PERFORM_SUBO_E_960(d, co, r1, r2, ci) \
+  ((d) = ((ci) & 2) != 0 ? (r1) - (r2) : (d), (co) = (ci))
+#define PERFORM_SUBO_GE_960(d, co, r1, r2, ci) \
+  ((d) = ((ci) & 3) != 0 ? (r1) - (r2) : (d), (co) = (ci))
+#define PERFORM_SUBO_L_960(d, co, r1, r2, ci) \
+  ((d) = ((ci) & 4) != 0 ? (r1) - (r2) : (d), (co) = (ci))
+#define PERFORM_SUBO_NE_960(d, co, r1, r2, ci) \
+  ((d) = ((ci) & 5) != 0 ? (r1) - (r2) : (d), (co) = (ci))
+#define PERFORM_SUBO_LE_960(d, co, r1, r2, ci) \
+  ((d) = ((ci) & 6) != 0 ? (r1) - (r2) : (d), (co) = (ci))
+#define PERFORM_SUBO_O_960(d, co, r1, r2, ci) \
+  ((d) = ((ci) & 7) != 0 ? (r1) - (r2) : (d), (co) = (ci))
+
+#define PERFORM_ALTERBIT(d, co, r1, r2, ci) \
+  ((d) = (EB1(ci) ? (r1) | ((word) 1 << TRUNC_CNT(r2))			\
+		  : (r1) & ~((word) 1 << TRUNC_CNT(r2))),		\
+   (co) = (ci))
+#define PERFORM_SETBIT(d, co, r1, r2, ci) \
+  ((d) = (r1) | ((word) 1 << TRUNC_CNT(r2)), (co) = (ci))
+#define PERFORM_CLRBIT(d, co, r1, r2, ci) \
+  ((d) = (r1) & ~((word) 1 << TRUNC_CNT(r2)), (co) = (ci))
+#define PERFORM_CHKBIT(d, co, r1, r2, ci) \
+  ((co) = IB1(((r1) >> TRUNC_CNT(r2)) & 1))
+#define PERFORM_NOTBIT(d, co, r1, r2, ci) \
+  ((d) = (r1) ^ ((word) 1 << TRUNC_CNT(r2)), (co) = (ci))
 
 enum goal_func
 {
@@ -1336,19 +1455,18 @@ enum prune_flags
   NULLIFYING_INSN = 8,
 };
 
-void
-synth(insn_t *sequence,
-      int n_insns,
-      word *values,
-      int n_values,
-      word desired_value,
-      int allowed_cost,
-      int cy_in,
-      int flags
 #if HAS_NULLIFICATION
-      , int nullication_flag
+void synth(insn_t *, int, word *, int, word, int, int, int, int);
+void synth_last(insn_t *, int, word *, int, word, int, int, int, int);
+void synth_nonskip(insn_t *, int, word *, int, word, int, int, int, int);
+void synth_nonskip_last(insn_t *, int, word *, int, word, int, int, int, int);
+void synth_condskip(insn_t *, int, word *, int, word, int, int, int, int);
+void synth_skip(insn_t *, int, word *, int, word, int, int, int, int);
+#else
+void synth(insn_t *, int, word *, int, word, int, int, int);
+void synth_last(insn_t *, int, word *, int, word, int, int, int);
 #endif
-      );
+
 void
 test_sequence(insn_t *sequence, int n_insns);
 int
